#!/usr/bin/env python
# Methods from SQLAlchemy utils modified.
# TODO: remove once https://github.com/kvesteri/sqlalchemy-utils/pull/372 is merged
#
###################################################################
#  SQLAlchemy utils methods
# Copyright (c) 2012, Konsta Vesterinen
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * The names of the contributors may not be used to endorse or promote products
# derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#  See https://github.com/kvesteri/sqlalchemy-utils/blob/master/LICENSE.
####################################################################
#
# Additional changes for project Thoth by Thoth team.
#

"""Utils for postgresql database."""

import os
import logging
from copy import copy

from sqlalchemy import create_engine
from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import OperationalError, ProgrammingError

_LOGGER = logging.getLogger(__name__)


def database_exists(url):
    """Check if a database exists.
    :param url: A SQLAlchemy engine URL.
    Performs backend-specific testing to quickly determine if a database
    exists on the server. ::
        database_exists('postgres://postgres@localhost/name')  #=> False
        create_database('postgres://postgres@localhost/name')
        database_exists('postgres://postgres@localhost/name')  #=> True
    Supports checking against a constructed URL as well. ::
        engine = create_engine('postgres://postgres@localhost/name')
        database_exists(engine.url)  #=> False
        create_database(engine.url)
        database_exists(engine.url)  #=> True

    Modified version from sqlalchemy_utils to take care of one issue.
    """

    def get_scalar_result(engine, sql):
        result_proxy = engine.execute(sql)
        result = result_proxy.scalar()
        result_proxy.close()
        engine.dispose()
        return result

    url = copy(make_url(url))
    database, url.database = url.database, os.getenv('KNOWLEDGE_GRAPH_DATABASE')
    engine = create_engine(url)

    if engine.dialect.name == "postgresql":
        text = "SELECT 1 FROM pg_database WHERE datname='%s'" % database
        return bool(get_scalar_result(engine, text))

    else:
        _LOGGER.exception("This implementation is valid only for postgresql dialects")


def create_database(url, encoding="utf8", template=None):
    """Issue the appropriate CREATE DATABASE statement.
    :param url: A SQLAlchemy engine URL.
    :param encoding: The encoding to create the database as.
    :param template:
        The name of the template from which to create the new database. At the
        moment only supported by PostgreSQL driver.
    To create a database, you can pass a simple URL that would have
    been passed to ``create_engine``. ::
        create_database('postgres://postgres@localhost/name')
    You may also pass the url from an existing engine. ::
        create_database(engine.url)
    Has full support for mysql, postgres, and sqlite. In theory,
    other database engines should be supported.

    Modified version from sqlalchemy_utils to take care of one issue.
    """

    url = copy(make_url(url))

    database, url.database = url.database, os.getenv('KNOWLEDGE_GRAPH_DATABASE')
    engine = create_engine(url)
    result_proxy = None

    if engine.dialect.name == "postgresql":
        if engine.driver == "psycopg2":
            from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT

            engine.raw_connection().set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)

        if not template:
            template = "template1"

        text = "CREATE DATABASE {0} ENCODING '{1}' TEMPLATE {2}".format(
            quote(engine, database), encoding, quote(engine, template)
        )
        result_proxy = engine.execute(text)

    else:
        _LOGGER.exception("This implementation is valid only for postgresql dialects")

    if result_proxy is not None:
        result_proxy.close()
    engine.dispose()



#!/usr/bin/env python3
# thoth-storages
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base and utilities for implementing SQLAlchemy based models."""

import logging
import datetime
from itertools import combinations
from typing import List

from sqlalchemy import Index
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import class_mapper
from sqlalchemy.orm import ColumnProperty
from sqlalchemy.exc import IntegrityError


Base = declarative_base()


_LOGGER = logging.getLogger(__name__)


class BaseExtension:
    """Extend base class with additional functionality."""

    @classmethod
    def get_or_create(cls, session, **kwargs):
        """Query for the given entity, create if it does not exist yet."""
        instance = session.query(cls).filter_by(**kwargs).first()
        if instance:
            return instance, True
        else:
            try:
                session.begin_nested()
                instance = cls(**kwargs)
                session.add(instance)
                session.commit()
                return instance, False
            except IntegrityError as exc:
                session.rollback()
                _LOGGER.warning(
                    "Integrity error on creating a new record; this can be due to "
                    "concurrent writes to database, recovering (attributes: %r): %s",
                    kwargs,
                    str(exc),
                )
                return session.query(cls).filter_by(**kwargs).one(), True

    @classmethod
    def attribute_names(cls):
        """Get names of attributes for the given model declaration."""
        return [prop.key for prop in class_mapper(cls).iterate_properties if isinstance(prop, ColumnProperty)]

    def to_dict(self, without_id: bool = True) -> dict:
        """Convert model to a dictionary representation keeping just rows as attributes."""
        result = {}
        for column in self.__table__.columns:
            if without_id and column.name == "id":
                continue

            if getattr(self, column.name) is None:
                result[column.name] = None
            else:
                value = getattr(self, column.name)
                if isinstance(value, (datetime.datetime, datetime.date)):
                    result[column.name] = str(value)
                else:
                    result[column.name] = value

        return result


def get_python_package_version_index_combinations() -> List[Index]:
    """Create index for all possible combinations which we can query."""
    result = []
    _columns_to_variate = (
        "os_name",
        "os_version",
        "python_version",
    )
    for i in range(0, len(_columns_to_variate) + 1):
        for j, variation in enumerate(combinations(_columns_to_variate, i)):
            result.append(
                Index(
                    f"python_package_version_index_idx_{i}{j}",
                    "package_name",
                    "package_version",
                    *variation,
                )
            )

    return result



"""Use Text instead of Strings all over the place

Revision ID: 0ea53ec16d6b
Revises: 125a71b49a50
Create Date: 2020-03-25 08:21:13.153294+00:00

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0ea53ec16d6b'
down_revision = '125a71b49a50'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('adviser_run', 'adviser_document_id',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('adviser_run', 'adviser_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('adviser_run', 'adviser_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('adviser_run', 'origin',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('adviser_run', 're_run_adviser_id',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('cve', 'advisory',
               existing_type=sa.VARCHAR(length=16384),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('cve', 'cve_id',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('cve', 'cve_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('cve', 'version_range',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('deb_dependency', 'package_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('deb_depends', 'version_range',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('deb_package_version', 'arch',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('deb_package_version', 'epoch',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('deb_package_version', 'package_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('deb_package_version', 'package_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('deb_pre_depends', 'version_range',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('deb_replaces', 'version_range',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('dependency_monkey_run', 'decision',
               existing_type=sa.VARCHAR(length=64),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('dependency_monkey_run', 'dependency_monkey_document_id',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('dependency_monkey_run', 'dependency_monkey_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('dependency_monkey_run', 'dependency_monkey_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('depends_on', 'version_range',
               existing_type=sa.VARCHAR(length=128),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('ecosystem_solver', 'ecosystem',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('ecosystem_solver', 'os_name',
               existing_type=sa.VARCHAR(length=128),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('ecosystem_solver', 'os_version',
               existing_type=sa.VARCHAR(length=8),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('ecosystem_solver', 'python_version',
               existing_type=sa.VARCHAR(length=8),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('ecosystem_solver', 'solver_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('ecosystem_solver', 'solver_version',
               existing_type=sa.VARCHAR(length=16),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('external_hardware_information', 'cpu_model_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('external_hardware_information', 'gpu_model_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('external_hardware_information', 'gpu_vendor',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'cuda_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'environment_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'image_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'image_sha',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'os_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'os_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'python_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('found_python_file', 'file',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('hardware_information', 'cpu_model_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('hardware_information', 'gpu_model_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('hardware_information', 'gpu_vendor',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('included_file', 'file',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('inspection_run', 'amun_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('inspection_run', 'inspection_document_id',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('package_analyzer_run', 'package_analysis_document_id',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('package_analyzer_run', 'package_analyzer_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('package_analyzer_run', 'package_analyzer_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('package_extract_run', 'analysis_document_id',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('package_extract_run', 'image_tag',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('package_extract_run', 'origin',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('package_extract_run', 'os_id',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('package_extract_run', 'os_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('package_extract_run', 'os_version_id',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('package_extract_run', 'package_extract_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('package_extract_run', 'package_extract_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('provenance_checker_run', 'origin',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('provenance_checker_run', 'provenance_checker_document_id',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('provenance_checker_run', 'provenance_checker_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('provenance_checker_run', 'provenance_checker_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('python_artifact', 'artifact_hash_sha256',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('python_artifact', 'artifact_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('python_file_digest', 'sha256',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('python_interpreter', 'link',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('python_interpreter', 'path',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('python_interpreter', 'version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('python_package_index', 'url',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('python_package_index', 'warehouse_api_url',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('python_package_requirement', 'name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('python_package_requirement', 'version_range',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('python_package_version', 'os_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('python_package_version', 'os_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('python_package_version', 'package_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('python_package_version', 'package_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('python_package_version', 'python_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('python_package_version_entity', 'package_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('python_package_version_entity', 'package_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('rpm_package_version', 'arch',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('rpm_package_version', 'epoch',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('rpm_package_version', 'package_identifier',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('rpm_package_version', 'package_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('rpm_package_version', 'package_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('rpm_package_version', 'release',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('rpm_requirement', 'rpm_requirement_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('software_environment', 'cuda_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('software_environment', 'environment_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('software_environment', 'image_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('software_environment', 'image_sha',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('software_environment', 'os_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('software_environment', 'os_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('software_environment', 'python_version',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=True)
    op.alter_column('solved', 'document_id',
               existing_type=sa.VARCHAR(length=128),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('versioned_symbol', 'library_name',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    op.alter_column('versioned_symbol', 'symbol',
               existing_type=sa.VARCHAR(length=256),
               type_=sa.Text(),
               existing_nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('versioned_symbol', 'symbol',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('versioned_symbol', 'library_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('solved', 'document_id',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=128),
               existing_nullable=False)
    op.alter_column('software_environment', 'python_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('software_environment', 'os_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('software_environment', 'os_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('software_environment', 'image_sha',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('software_environment', 'image_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('software_environment', 'environment_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('software_environment', 'cuda_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('rpm_requirement', 'rpm_requirement_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('rpm_package_version', 'release',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('rpm_package_version', 'package_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('rpm_package_version', 'package_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('rpm_package_version', 'package_identifier',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('rpm_package_version', 'epoch',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('rpm_package_version', 'arch',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('python_package_version_entity', 'package_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('python_package_version_entity', 'package_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('python_package_version', 'python_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('python_package_version', 'package_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('python_package_version', 'package_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('python_package_version', 'os_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('python_package_version', 'os_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('python_package_requirement', 'version_range',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('python_package_requirement', 'name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('python_package_index', 'warehouse_api_url',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('python_package_index', 'url',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('python_interpreter', 'version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('python_interpreter', 'path',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('python_interpreter', 'link',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('python_file_digest', 'sha256',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('python_artifact', 'artifact_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('python_artifact', 'artifact_hash_sha256',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('provenance_checker_run', 'provenance_checker_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('provenance_checker_run', 'provenance_checker_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('provenance_checker_run', 'provenance_checker_document_id',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('provenance_checker_run', 'origin',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('package_extract_run', 'package_extract_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('package_extract_run', 'package_extract_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('package_extract_run', 'os_version_id',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('package_extract_run', 'os_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('package_extract_run', 'os_id',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('package_extract_run', 'origin',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('package_extract_run', 'image_tag',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('package_extract_run', 'analysis_document_id',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('package_analyzer_run', 'package_analyzer_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('package_analyzer_run', 'package_analyzer_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('package_analyzer_run', 'package_analysis_document_id',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('inspection_run', 'inspection_document_id',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('inspection_run', 'amun_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('included_file', 'file',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('hardware_information', 'gpu_vendor',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('hardware_information', 'gpu_model_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('hardware_information', 'cpu_model_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('found_python_file', 'file',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('external_software_environment', 'python_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'os_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'os_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'image_sha',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'image_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'environment_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('external_software_environment', 'cuda_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('external_hardware_information', 'gpu_vendor',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('external_hardware_information', 'gpu_model_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('external_hardware_information', 'cpu_model_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('ecosystem_solver', 'solver_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=16),
               existing_nullable=False)
    op.alter_column('ecosystem_solver', 'solver_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('ecosystem_solver', 'python_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=8),
               existing_nullable=False)
    op.alter_column('ecosystem_solver', 'os_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=8),
               existing_nullable=False)
    op.alter_column('ecosystem_solver', 'os_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=128),
               existing_nullable=False)
    op.alter_column('ecosystem_solver', 'ecosystem',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('depends_on', 'version_range',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=128),
               existing_nullable=True)
    op.alter_column('dependency_monkey_run', 'dependency_monkey_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('dependency_monkey_run', 'dependency_monkey_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('dependency_monkey_run', 'dependency_monkey_document_id',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('dependency_monkey_run', 'decision',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=64),
               existing_nullable=False)
    op.alter_column('deb_replaces', 'version_range',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('deb_pre_depends', 'version_range',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('deb_package_version', 'package_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('deb_package_version', 'package_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('deb_package_version', 'epoch',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('deb_package_version', 'arch',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('deb_depends', 'version_range',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('deb_dependency', 'package_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('cve', 'version_range',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('cve', 'cve_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('cve', 'cve_id',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('cve', 'advisory',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=16384),
               existing_nullable=True)
    op.alter_column('adviser_run', 're_run_adviser_id',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('adviser_run', 'origin',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=True)
    op.alter_column('adviser_run', 'adviser_version',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('adviser_run', 'adviser_name',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    op.alter_column('adviser_run', 'adviser_document_id',
               existing_type=sa.Text(),
               type_=sa.VARCHAR(length=256),
               existing_nullable=False)
    # ### end Alembic commands ###



"""Introduce PI PyBench for Python Interpreter

Revision ID: f00fc22d1589
Revises: 3d0f03348c25
Create Date: 2019-12-16 12:40:36.344874+00:00

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f00fc22d1589'
down_revision = '3d0f03348c25'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('pi_pybench',
    sa.Column('component', sa.String(length=256), nullable=False),
    sa.Column('origin', sa.String(length=256), nullable=False),
    sa.Column('version', sa.String(length=256), nullable=False),
    sa.Column('overall_score', sa.Float(), nullable=True),
    sa.Column('exit_code', sa.Integer(), nullable=False),
    sa.Column('ru_utime', sa.Float(), nullable=False),
    sa.Column('ru_stime', sa.Float(), nullable=False),
    sa.Column('ru_maxrss', sa.Integer(), nullable=False),
    sa.Column('ru_ixrss', sa.Integer(), nullable=False),
    sa.Column('ru_idrss', sa.Integer(), nullable=False),
    sa.Column('ru_isrss', sa.Integer(), nullable=False),
    sa.Column('ru_minflt', sa.Integer(), nullable=False),
    sa.Column('ru_majflt', sa.Integer(), nullable=False),
    sa.Column('ru_nswap', sa.Integer(), nullable=False),
    sa.Column('ru_inblock', sa.Integer(), nullable=False),
    sa.Column('ru_oublock', sa.Integer(), nullable=False),
    sa.Column('ru_msgsnd', sa.Integer(), nullable=False),
    sa.Column('ru_msgrcv', sa.Integer(), nullable=False),
    sa.Column('ru_nsignals', sa.Integer(), nullable=False),
    sa.Column('ru_nvcsw', sa.Integer(), nullable=False),
    sa.Column('ru_nivcsw', sa.Integer(), nullable=False),
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('inspection_run_id', sa.Integer(), nullable=False),
    sa.Column('rounds', sa.Integer(), nullable=False),
    sa.Column('built_in_function_calls_average', sa.Float(), nullable=False),
    sa.Column('built_in_method_lookup_average', sa.Float(), nullable=False),
    sa.Column('compare_floats_average', sa.Float(), nullable=False),
    sa.Column('compare_floats_integers_average', sa.Float(), nullable=False),
    sa.Column('compare_integers_average', sa.Float(), nullable=False),
    sa.Column('compare_interned_strings_average', sa.Float(), nullable=False),
    sa.Column('compare_longs_average', sa.Float(), nullable=False),
    sa.Column('compare_strings_average', sa.Float(), nullable=False),
    sa.Column('compare_unicode_average', sa.Float(), nullable=False),
    sa.Column('concat_strings_average', sa.Float(), nullable=False),
    sa.Column('concat_unicode_average', sa.Float(), nullable=False),
    sa.Column('create_instances_average', sa.Float(), nullable=False),
    sa.Column('create_new_instances_average', sa.Float(), nullable=False),
    sa.Column('create_strings_with_concat_average', sa.Float(), nullable=False),
    sa.Column('create_unicode_with_concat_average', sa.Float(), nullable=False),
    sa.Column('dict_creation_average', sa.Float(), nullable=False),
    sa.Column('dict_with_float_keys_average', sa.Float(), nullable=False),
    sa.Column('dict_with_integer_keys_average', sa.Float(), nullable=False),
    sa.Column('dict_with_string_keys_average', sa.Float(), nullable=False),
    sa.Column('for_loops_average', sa.Float(), nullable=False),
    sa.Column('if_then_else_average', sa.Float(), nullable=False),
    sa.Column('list_slicing_average', sa.Float(), nullable=False),
    sa.Column('nested_for_loops_average', sa.Float(), nullable=False),
    sa.Column('normal_class_attribute_average', sa.Float(), nullable=False),
    sa.Column('normal_instance_attribute_average', sa.Float(), nullable=False),
    sa.Column('python_function_calls_average', sa.Float(), nullable=False),
    sa.Column('python_method_calls_average', sa.Float(), nullable=False),
    sa.Column('recursion_average', sa.Float(), nullable=False),
    sa.Column('second_import_average', sa.Float(), nullable=False),
    sa.Column('second_package_import_average', sa.Float(), nullable=False),
    sa.Column('second_submodule_import_average', sa.Float(), nullable=False),
    sa.Column('simple_complex_arithmetic_average', sa.Float(), nullable=False),
    sa.Column('simple_dict_manipulation_average', sa.Float(), nullable=False),
    sa.Column('simple_float_arithmetic_average', sa.Float(), nullable=False),
    sa.Column('simple_int_float_arithmetic_average', sa.Float(), nullable=False),
    sa.Column('simple_integer_arithmetic_average', sa.Float(), nullable=False),
    sa.Column('simple_list_manipulation_average', sa.Float(), nullable=False),
    sa.Column('simple_long_arithmetic_average', sa.Float(), nullable=False),
    sa.Column('small_lists_average', sa.Float(), nullable=False),
    sa.Column('small_tuples_average', sa.Float(), nullable=False),
    sa.Column('special_class_attribute_average', sa.Float(), nullable=False),
    sa.Column('special_instance_attribute_average', sa.Float(), nullable=False),
    sa.Column('string_mappings_average', sa.Float(), nullable=False),
    sa.Column('string_predicates_average', sa.Float(), nullable=False),
    sa.Column('string_slicing_average', sa.Float(), nullable=False),
    sa.Column('try_except_average', sa.Float(), nullable=False),
    sa.Column('try_raise_except_average', sa.Float(), nullable=False),
    sa.Column('tuple_slicing_average', sa.Float(), nullable=False),
    sa.Column('unicode_mappings_average', sa.Float(), nullable=False),
    sa.Column('unicode_predicates_average', sa.Float(), nullable=False),
    sa.Column('unicode_properties_average', sa.Float(), nullable=False),
    sa.Column('unicode_slicing_average', sa.Float(), nullable=False),
    sa.Column('totals_average', sa.Float(), nullable=False),
    sa.ForeignKeyConstraint(['inspection_run_id'], ['inspection_run.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.add_column('pi_conv1d', sa.Column('component', sa.String(length=256), nullable=False))
    op.alter_column('pi_conv1d', 'origin',
               existing_type=sa.VARCHAR(length=256),
               nullable=False)
    op.drop_column('pi_conv1d', 'framework')
    op.add_column('pi_conv2d', sa.Column('component', sa.String(length=256), nullable=False))
    op.alter_column('pi_conv2d', 'origin',
               existing_type=sa.VARCHAR(length=256),
               nullable=False)
    op.drop_column('pi_conv2d', 'framework')
    op.add_column('pi_matmul', sa.Column('component', sa.String(length=256), nullable=False))
    op.alter_column('pi_matmul', 'origin',
               existing_type=sa.VARCHAR(length=256),
               nullable=False)
    op.drop_column('pi_matmul', 'framework')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('pi_matmul', sa.Column('framework', sa.VARCHAR(length=256), autoincrement=False, nullable=True))
    op.alter_column('pi_matmul', 'origin',
               existing_type=sa.VARCHAR(length=256),
               nullable=True)
    op.drop_column('pi_matmul', 'component')
    op.add_column('pi_conv2d', sa.Column('framework', sa.VARCHAR(length=256), autoincrement=False, nullable=True))
    op.alter_column('pi_conv2d', 'origin',
               existing_type=sa.VARCHAR(length=256),
               nullable=True)
    op.drop_column('pi_conv2d', 'component')
    op.add_column('pi_conv1d', sa.Column('framework', sa.VARCHAR(length=256), autoincrement=False, nullable=True))
    op.alter_column('pi_conv1d', 'origin',
               existing_type=sa.VARCHAR(length=256),
               nullable=True)
    op.drop_column('pi_conv1d', 'component')
    op.drop_table('pi_pybench')
    # ### end Alembic commands ###



#!/usr/bin/env python3
# thoth-storages
# Copyright(C) 2019, 2020 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Performance indicators (models) used in Thoth with their schemas."""

from functools import partial
import logging

from typing import Dict, Any

from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import ForeignKey
from sqlalchemy import String
from sqlalchemy import Float
from sqlalchemy.orm import relationship
from sqlalchemy.orm import Session

from .models_base import BaseExtension
from .models_base import Base

_LOGGER = logging.getLogger(__name__)


class PerformanceIndicatorBase:
    """A base class for implementing performance indicators."""

    # Component used for the performance indicator.
    component = Column(String(256), nullable=False)

    # Origin from where the performance indicator was obtained. In case of Git repo,
    # it holds Git repo URL, in case of URL it holds URL to the script.
    origin = Column(String(256), nullable=False)

    # Reference of the script, in case of Git repo it holds commit SHA, in case of URL it carries
    # SHA256 of the script which was used to test the performance with (performance indicator script).
    version = Column(String(256), nullable=False)

    # This one is used later on in queries in adviser, all the relevant performance indicators should
    # respect this property and place results in there.
    overall_score = Column(Float, nullable=True)

    # The actual exit code of the performance indicator.
    exit_code = Column(Integer, nullable=False)

    # Process statistics:
    #   https://docs.python.org/3/library/resource.html#resource.getrusage
    ru_utime = Column(Float, nullable=False)
    ru_stime = Column(Float, nullable=False)
    ru_maxrss = Column(Integer, nullable=False)
    ru_ixrss = Column(Integer, nullable=False)
    ru_idrss = Column(Integer, nullable=False)
    ru_isrss = Column(Integer, nullable=False)
    ru_minflt = Column(Integer, nullable=False)
    ru_majflt = Column(Integer, nullable=False)
    ru_nswap = Column(Integer, nullable=False)
    ru_inblock = Column(Integer, nullable=False)
    ru_oublock = Column(Integer, nullable=False)
    ru_msgsnd = Column(Integer, nullable=False)
    ru_msgrcv = Column(Integer, nullable=False)
    ru_nsignals = Column(Integer, nullable=False)
    ru_nvcsw = Column(Integer, nullable=False)
    ru_nivcsw = Column(Integer, nullable=False)

    @classmethod
    def create_from_report(
        cls, session: Session, inspection_specification: Dict[str, Any], inspection_result: Dict[str, Any], inspection_run_id: int
    ) -> "PerformanceIndicatorBase":
        """Create performance indicator record together with related observed performance edge based on inspection."""
        # Place core parts of the base class into the model.
        overall_score = inspection_result.get("overall_score")
        if overall_score is None:
            _LOGGER.warning("No overall score detected in performance indicator %r", overall_score)

        partial_model = partial(
            cls.get_or_create,
            session,
            inspection_run_id=inspection_run_id,
            component=inspection_result.get("component"),
            origin=inspection_specification["script"],
            version=inspection_result.get("version")
            or inspection_result["script_sha256"],
            overall_score=overall_score,
            exit_code=inspection_result.get("exit_code"),
            ru_utime=inspection_result.get("usage", {}).get("ru_utime"),
            ru_stime=inspection_result.get("usage", {}).get("ru_stime"),
            ru_maxrss=inspection_result.get("usage", {}).get("ru_maxrss"),
            ru_ixrss=inspection_result.get("usage", {}).get("ru_ixrss"),
            ru_idrss=inspection_result.get("usage", {}).get("ru_idrss"),
            ru_isrss=inspection_result.get("usage", {}).get("ru_isrss"),
            ru_minflt=inspection_result.get("usage", {}).get("ru_minflt"),
            ru_majflt=inspection_result.get("usage", {}).get("ru_majflt"),
            ru_nswap=inspection_result.get("usage", {}).get("ru_nswap"),
            ru_inblock=inspection_result.get("usage", {}).get("ru_inblock"),
            ru_oublock=inspection_result.get("usage", {}).get("ru_oublock"),
            ru_msgsnd=inspection_result.get("usage", {}).get("ru_msgsnd"),
            ru_msgrcv=inspection_result.get("usage", {}).get("ru_msgrcv"),
            ru_nsignals=inspection_result.get("usage", {}).get("ru_nsignals"),
            ru_nvcsw=inspection_result.get("usage", {}).get("ru_nvcsw"),
            ru_nivcsw=inspection_result.get("usage", {}).get("ru_nivcsw"),
        )

        return cls.from_report(inspection_result, partial_model)

    @classmethod
    def from_report(cls, inspection_result: Dict[str, Any], partial_model: type(partial)) -> "PerformanceIndicatorBase":
        """Create model from the inspection report respecting parameters and result reported by the indicator."""
        kwargs = {}
        for parameter, parameter_value in inspection_result["@parameters"].items():
            kwargs[parameter] = parameter_value

        for result_name, result_value in inspection_result["@result"].items():
            if result_name in kwargs:
                raise ValueError("Collision in result name and parameter name")

            kwargs[result_name] = result_value

        return partial_model(**kwargs)


class PiMatmul(Base, BaseExtension, PerformanceIndicatorBase):
    """A class for representing a matrix multiplication micro-performance test."""

    __tablename__ = "pi_matmul"

    id = Column(Integer, primary_key=True, autoincrement=True)

    inspection_run_id = Column(Integer, ForeignKey("inspection_run.id"), nullable=False)
    inspection_run = relationship("InspectionRun", back_populates="matmul_perf_indicators")

    # Device used during performance indicator run - CPU/GPU/TPU/...
    device = Column(String(256), nullable=False)

    # Size of the matrix tested.
    matrix_size = Column(Integer, nullable=False)

    # Type of item in the matrix.
    dtype = Column(String(256), nullable=False)

    # Number of repetitions of matrix multiplication performed.
    reps = Column(Integer, nullable=False)

    # Elapsed seconds.
    elapsed = Column(Float, nullable=False)

    # Final rate gflops/s.
    rate = Column(Float, nullable=False)


class PiConv1D(Base, BaseExtension, PerformanceIndicatorBase):
    """A class for representing a conv1D micro-performance test."""

    __tablename__ = "pi_conv1d"

    id = Column(Integer, primary_key=True, autoincrement=True)

    inspection_run_id = Column(Integer, ForeignKey("inspection_run.id"), nullable=False)
    inspection_run = relationship("InspectionRun", back_populates="conv1d_perf_indicators")

    # Device used during performance indicator run - CPU/GPU/TPU/...
    device = Column(String(256), nullable=False)

    # Type of item in the tensor.
    dtype = Column(String(256), nullable=False)

    # Number of repetitions of conv2d performed.
    reps = Column(Integer, nullable=False)

    # Data format NHWC Channel_last or NCHW Channel_first
    data_format = Column(String(256), nullable=False)

    # INPUT TENSOR
    batch = Column(Integer, nullable=False)
    input_width = Column(Integer, nullable=False)
    input_channels = Column(Integer, nullable=False)

    # FILTER
    filter_width = Column(Integer, nullable=False)
    output_channels = Column(Integer, nullable=False)

    # Stride, the speed by which the filter moves across the image
    strides = Column(Integer, nullable=False)

    # Padding
    padding = Column(Integer, nullable=False)

    # Elapsed seconds.
    elapsed = Column(Float, nullable=False)

    # Final rate gflops/s.
    rate = Column(Float, nullable=False)


class PiConv2D(Base, BaseExtension, PerformanceIndicatorBase):
    """A class for representing a conv2D micro-performance test."""

    __tablename__ = "pi_conv2d"

    id = Column(Integer, primary_key=True, autoincrement=True)

    inspection_run_id = Column(Integer, ForeignKey("inspection_run.id"), nullable=False)
    inspection_run = relationship("InspectionRun", back_populates="conv2d_perf_indicators")

    # Device used during performance indicator run - CPU/GPU/TPU/...
    device = Column(String(256), nullable=False)

    # Type of item in the tensor.
    dtype = Column(String(256), nullable=False)

    # Number of repetitions of conv2d performed.
    reps = Column(Integer, nullable=False)

    # Data format NHWC Channel_last or NCHW Channel_first
    data_format = Column(String(256), nullable=False)

    # INPUT TENSOR
    batch = Column(Integer, nullable=False)
    input_height = Column(Integer, nullable=False)
    input_width = Column(Integer, nullable=False)
    input_channels = Column(Integer, nullable=False)

    # FILTER
    filter_height = Column(Integer, nullable=False)
    filter_width = Column(Integer, nullable=False)
    output_channels = Column(Integer, nullable=False)

    # Stride, the speed by which the filter moves across the image
    strides = Column(Integer, nullable=False)

    # Padding
    padding = Column(Integer, nullable=False)

    # Elapsed seconds.
    elapsed = Column(Float, nullable=False)

    # Final rate gflops/s.
    rate = Column(Float, nullable=False)


class PiPyBench(Base, BaseExtension, PerformanceIndicatorBase):
    """A class for representing Pybench results for Python Interpreter."""

    __tablename__ = "pi_pybench"

    id = Column(Integer, primary_key=True, autoincrement=True)

    inspection_run_id = Column(Integer, ForeignKey("inspection_run.id"), nullable=False)
    inspection_run = relationship("InspectionRun", back_populates="pybench_perf_indicators")

    # Number of rounds used for each function.
    rounds = Column(Integer, nullable=False)

    # Average results for each function
    built_in_function_calls_average = Column(Float, nullable=False)
    built_in_method_lookup_average = Column(Float, nullable=False)
    compare_floats_average = Column(Float, nullable=False)
    compare_floats_integers_average = Column(Float, nullable=False)
    compare_integers_average = Column(Float, nullable=False)
    compare_interned_strings_average = Column(Float, nullable=False)
    compare_longs_average = Column(Float, nullable=False)
    compare_strings_average = Column(Float, nullable=False)
    compare_unicode_average = Column(Float, nullable=False)
    concat_strings_average = Column(Float, nullable=False)
    concat_unicode_average = Column(Float, nullable=False)
    create_instances_average = Column(Float, nullable=False)
    create_new_instances_average = Column(Float, nullable=False)
    create_strings_with_concat_average = Column(Float, nullable=False)
    create_unicode_with_concat_average = Column(Float, nullable=False)
    dict_creation_average = Column(Float, nullable=False)
    dict_with_float_keys_average = Column(Float, nullable=False)
    dict_with_integer_keys_average = Column(Float, nullable=False)
    dict_with_string_keys_average = Column(Float, nullable=False)
    for_loops_average = Column(Float, nullable=False)
    if_then_else_average = Column(Float, nullable=False)
    list_slicing_average = Column(Float, nullable=False)
    nested_for_loops_average = Column(Float, nullable=False)
    normal_class_attribute_average = Column(Float, nullable=False)
    normal_instance_attribute_average = Column(Float, nullable=False)
    python_function_calls_average = Column(Float, nullable=False)
    python_method_calls_average = Column(Float, nullable=False)
    recursion_average = Column(Float, nullable=False)
    second_import_average = Column(Float, nullable=False)
    second_package_import_average = Column(Float, nullable=False)
    second_submodule_import_average = Column(Float, nullable=False)
    simple_complex_arithmetic_average = Column(Float, nullable=False)
    simple_dict_manipulation_average = Column(Float, nullable=False)
    simple_float_arithmetic_average = Column(Float, nullable=False)
    simple_int_float_arithmetic_average = Column(Float, nullable=False)
    simple_integer_arithmetic_average = Column(Float, nullable=False)
    simple_list_manipulation_average = Column(Float, nullable=False)
    simple_long_arithmetic_average = Column(Float, nullable=False)
    small_lists_average = Column(Float, nullable=False)
    small_tuples_average = Column(Float, nullable=False)
    special_class_attribute_average = Column(Float, nullable=False)
    special_instance_attribute_average = Column(Float, nullable=False)
    string_mappings_average = Column(Float, nullable=False)
    string_predicates_average = Column(Float, nullable=False)
    string_slicing_average = Column(Float, nullable=False)
    try_except_average = Column(Float, nullable=False)
    try_raise_except_average = Column(Float, nullable=False)
    tuple_slicing_average = Column(Float, nullable=False)
    unicode_mappings_average = Column(Float, nullable=False)
    unicode_predicates_average = Column(Float, nullable=False)
    unicode_properties_average = Column(Float, nullable=False)
    unicode_slicing_average = Column(Float, nullable=False)

    # Total average results
    totals_average = Column(Float, nullable=False)


PERFORMANCE_MODELS_ML_FRAMEWORKS = frozenset((PiMatmul, PiConv1D, PiConv2D))

ALL_PERFORMANCE_MODELS = frozenset((PiMatmul, PiConv1D, PiConv2D, PiPyBench))

PERFORMANCE_MODEL_BY_NAME = {model_class.__name__: model_class for model_class in ALL_PERFORMANCE_MODELS}



"""Added new table to link adviser run and python package version entity

Revision ID: 74254b596f00
Revises: ed56455efd82
Create Date: 2020-03-20 15:31:37.608247+00:00

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '74254b596f00'
down_revision = 'ed56455efd82'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('has_unresolved',
    sa.Column('adviser_run_id', sa.Integer(), nullable=False),
    sa.Column('python_package_version_entity_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['adviser_run_id'], ['adviser_run.id'], ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['python_package_version_entity_id'], ['python_package_version_entity.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('adviser_run_id', 'python_package_version_entity_id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('has_unresolved')
    # ### end Alembic commands ###



"""Add flag for missing package version

Revision ID: ed56455efd82
Revises: 0c8eab5d4197
Create Date: 2020-02-19 19:02:38.296829+00:00

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ed56455efd82'
down_revision = '0c8eab5d4197'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('python_package_version', sa.Column('is_missing', sa.Boolean(), nullable=False, server_default="false"))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('python_package_version', 'is_missing')
    # ### end Alembic commands ###



"""Add indexes needed for revsolver

Revision ID: a7442899ef28
Revises: ecb0ca258835
Create Date: 2020-04-21 19:19:43.848258+00:00

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'a7442899ef28'
down_revision = 'ecb0ca258835'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('depends_on_entity_id_idx', 'depends_on', ['entity_id'], unique=False)
    op.create_index('python_package_version_environment_idx', 'python_package_version', ['os_name', 'os_version', 'python_version'], unique=False)
    op.create_index('python_package_version_entity_package_name_idx', 'python_package_version_entity', ['package_name'], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('python_package_version_entity_package_name_idx', table_name='python_package_version_entity')
    op.drop_index('python_package_version_environment_idx', table_name='python_package_version')
    op.drop_index('depends_on_entity_id_idx', table_name='depends_on')
    # ### end Alembic commands ###



"""Added a relation between PythonPackageVersion and SIAggregated

Revision ID: 84588707e6ba
Revises: 83e2900c3721
Create Date: 2020-08-26 18:55:20.736024+00:00

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '84588707e6ba'
down_revision = '83e2900c3721'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('si_aggregated', sa.Column('python_package_version_id', sa.Integer(), nullable=False))
    op.create_foreign_key(None, 'si_aggregated', 'python_package_version', ['python_package_version_id'], ['id'], ondelete='CASCADE')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, 'si_aggregated', type_='foreignkey')
    op.drop_column('si_aggregated', 'python_package_version_id')
    # ### end Alembic commands ###



"""Create index for get_depends_query.

Revision ID: 79e1d320db3d
Revises: 8231fee9a210
Create Date: 2019-11-29 16:59:44.869319+00:00

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '79e1d320db3d'
down_revision = '8231fee9a210'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index('depends_on_version_id_idx', 'depends_on', ['version_id'], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('depends_on_version_id_idx', table_name='depends_on')
    # ### end Alembic commands ###
